CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                     C
C                            Pio (Fortran)                            C
C                                                                     C
C                    Polyglot Input/Output Library                    C
C                                                                     C
C               EPI Pomdapi, INRIA Paris - Rocquencourt               C
C                                                                     C
C  Copyright 2005-2014 INRIA.                                         C
C  Institut National de Recherche en Informatique et en Automatique.  C
C  All rights reserved. This file is distributed only by permission.  C
C                                                                     C
C  Francois Clement <Francois.Clement@inria.fr>                       C
C  Pierre Weis <Pierre.Weis@inria.fr>                                 C
C  Hend Ben Ameur <hbenameur@yahoo.ca>                                C
C                                                                     C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

C $Id: memory_mapping.fpp,v 1.2 2014/01/10 19:22:10 fclement Exp $

C     
C     Routines to handle the mapping of the low-level memory.
C     
C     The mapping describes where and how values are stored into
C     the low-level memory
C     
      
#include "pio_constants.h"
      
      subroutine invalidate_memory_mapping(sbr, io)
C     Invalidate the memory mapping.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io
      
C     Local variables.
      logical b
      integer i
      
C     External function.
      logical pio_get_debug_mode, get_memory_mapping_clean
      
      call set_memory_mapping_clean(io, .False.)
      
      call store_communication_type(sbr, io, INVALID_COMMUNICATION_TYPE)
      call store_communication_values_number(sbr, io, INVALID_VALUES_NUMBER)
      call store_communication_name_begin(sbr, io, INVALID_NAME_BEGIN)
      call store_communication_name_end(sbr, io, INVALID_NAME_END)
      call store_communication_begin(sbr, io, INVALID_COMMUNICATION_BEGIN)
      call store_communication_end(sbr, io, INVALID_COMMUNICATION_END)
      
      do i = SLOT_VALUE_DESCRIPTOR_ADDRESSES, MAX_MEMORY_MAPPING_LOCATION
         call rand_store_slot(sbr, io, i, INVALID_VALUES)
      enddo
      
      fetch_value_number(io) = INVALID_FETCH_VALUE_NUMBER
      store_value_number(io) = INVALID_STORE_VALUE_NUMBER
      
      fetch_value_descriptor_ptr(io) = INVALID_FETCH_VALUE_DESCRIPTOR_PTR
      fetch_value_descriptor_lim(io) = INVALID_FETCH_VALUE_DESCRIPTOR_LIM
      store_value_descriptor_ptr(io) = INVALID_STORE_VALUE_DESCRIPTOR_PTR
      
      if (pio_get_debug_mode()) then
C     Remember the memory mapping clean state before allowing reading
C     permission.
         b = get_memory_mapping_clean(io)
         call set_memory_mapping_clean(io, .True.)
         call pio_write_memory_mapping(io)
         call set_memory_mapping_clean(io, b)
      endif
      
      return
      end
      
      
      subroutine check_value_number_invariant(sbr, io)
C     Check the invariant on value numbers.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io
      
C     Local variables.
      logical out_of_bounds, unwritten_data_access, overflow
      integer fvn, svn
      
      fvn = fetch_value_number(io)
      svn = store_value_number(io)
      
      out_of_bounds = fvn .lt. FIRST_VALUE_NUMBER
      unwritten_data_access = fvn .gt. svn
      overflow = svn .gt. MAX_VALUE_NUMBER
      
      if (out_of_bounds) then
         call error_out_of_bounds(sbr, io, 'fetch value number')
      endif
      
      if (unwritten_data_access) then
         call error_unwritten_data_access
     +        (sbr, io, 'value number', 'fetch value number', svn, fvn)
      endif
      
      if (overflow) then
         call error_overflow(sbr, io, 'value number')
      endif
      
      return
      end
      
      
      subroutine init_memory_mapping(sbr, io)
C     Initialize the memory mapping.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io
      
      call invalidate_memory_mapping(sbr, io)
      
      fetch_value_number(io) = FIRST_VALUE_NUMBER
      store_value_number(io) = FIRST_VALUE_NUMBER
      
      call check_value_number_invariant(sbr, io)
      
      fetch_value_descriptor_ptr(io) = FIRST_POINTER
      fetch_value_descriptor_lim(io) = FIRST_POINTER
      store_value_descriptor_ptr(io) = FIRST_POINTER
      
      call check_value_descriptor_invariant(sbr, io)
      
      return
      end
      
      
      integer function get_store_value_number(io)
C     Get the store value number.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input argument.
      integer io
      
      get_store_value_number = store_value_number(io)
      
      return
      end
      
      
      subroutine set_store_value_number(sbr, io, n)
C     Set the store value number to n.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, n
      
C     External function.
      logical pio_get_debug_mode
      
      store_value_number(io) = n
      
      call check_value_number_invariant(sbr, io)
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *)
     +        '[Pio/Fortran] Setting store_value_number to ', n
      endif
      
      return
      end
      
      
      subroutine incr_store_value_number(sbr, io)
C     Increment the store value number.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io
      
      call set_store_value_number(sbr, io, store_value_number(io) + 1)
      
      return
      end
      
      
      integer function get_fetch_value_number(io)
C     Get the fetch value number.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input argument.
      integer io
      
      get_fetch_value_number = fetch_value_number(io)
      
      return
      end
      
      
      subroutine set_fetch_value_number(sbr, io, n)
C     Set the fetch value number to n.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, n
      
C     External function.
      logical pio_get_debug_mode
      
      fetch_value_number(io) = n
      
      call check_value_number_invariant(sbr, io)
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *)
     +        '[Pio/Fortran] Setting fetch_value_number to ', n
      endif
      
      return
      end
      
      
      subroutine incr_fetch_value_number(sbr, io)
C     Increments the fetch value number.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io
      
C     Local variable.
      integer n
      
      n = fetch_value_number(io) + 1
      
      call set_fetch_value_number(sbr, io, n)
      
      return
      end
      
      
      subroutine check_value_descriptor_invariant(sbr, io)
C     Check the invariant on value descriptor pointers.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io
      
C     Local variables.
      logical out_of_bounds, invalid_data_access
      logical unwritten_data_access, overflow
      integer fvd_ptr, fvd_lim, svd_ptr
      
      fvd_ptr = fetch_value_descriptor_ptr(io)
      fvd_lim = fetch_value_descriptor_lim(io)
      svd_ptr = store_value_descriptor_ptr(io)
      
      out_of_bounds = fvd_ptr .lt. MIN_MEMORY_MAPPING_LOCATION
      invalid_data_access = fvd_ptr .gt. fvd_lim
      unwritten_data_access = fvd_lim .gt. svd_ptr
      overflow = svd_ptr .gt. MAX_MEMORY_MAPPING_LOCATION
      
      if (out_of_bounds) then
         call error_out_of_bounds
     +        (sbr, io, 'fetch value descriptor pointer')
      endif
      
      if (invalid_data_access) then
         call error_invalid_data_access
     +        (sbr, io, 'value descriptor',
     +        'fetch value descriptor pointer', fvd_lim, fvd_ptr)
      endif
      
      if (unwritten_data_access) then
         call error_unwritten_data_access
     +        (sbr, io, 'value descriptor',
     +        'fetch value descriptor pointer', svd_ptr, fvd_lim)
      endif
      
      if (overflow) then
         call error_overflow(sbr, io, 'memory mapping')
      endif
      
      return
      end
      
      
      integer function get_store_value_descriptor_ptr(io)
C     Get the store value descriptor pointer.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input argument.
      integer io
      
      get_store_value_descriptor_ptr = store_value_descriptor_ptr(io)
      
      return
      end
      
      
      subroutine set_store_value_descriptor_ptr(sbr, io, ptr)
C     Set the store value descriptor pointer to ptr.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, ptr
      
C     External function.
      logical pio_get_debug_mode
      
      store_value_descriptor_ptr(io) = ptr
      
      call check_value_descriptor_invariant(sbr, io)
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *)
     +        '[Pio/Fortran] Setting store_value_descriptor_ptr to ',
     +     ptr
      endif
      
      return
      end
      
      
      subroutine advance_store_value_descriptor_ptr(sbr, io, len)
C     Advance the store value descriptor pointer of len locations.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, len
      
      call set_store_value_descriptor_ptr
     +     (sbr, io, store_value_descriptor_ptr(io) + len)
      
      return
      end
      
      
      subroutine incr_store_value_descriptor_ptr(sbr, io)
C     Increment the store value descriptor pointer.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io
      
      call advance_store_value_descriptor_ptr(sbr, io, 1)
      
      return
      end
      
      
      integer function get_fetch_value_descriptor_ptr(io)
C     Get the fetch value descriptor pointer.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input argument.
      integer io
      
      get_fetch_value_descriptor_ptr = fetch_value_descriptor_ptr(io)
      
      return
      end
      
      
      subroutine set_fetch_value_descriptor_ptr(sbr, io, ptr)
C     Set the fetch value descriptor pointer to ptr.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, ptr
      
C     External function.
      logical pio_get_debug_mode
      
      fetch_value_descriptor_ptr(io) = ptr
      
      call check_value_descriptor_invariant(sbr, io)
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *)
     +        '[Pio/Fortran] Setting fetch_value_descriptor_ptr to ',
     +     ptr
      endif
      
      return
      end
      
      
      subroutine incr_fetch_value_descriptor_ptr(sbr, io)
C     Increment the fetch value descriptor pointer.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io
      
      call set_fetch_value_descriptor_ptr
     +     (sbr, io, fetch_value_descriptor_ptr(io) + 1)
      
      return
      end
      
      
      integer function get_fetch_value_descriptor_lim(io)
C     Get the fetch value descriptor limit.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input argument.
      integer io
      
      get_fetch_value_descriptor_lim = fetch_value_descriptor_lim(io)
      
      return
      end
      
      
      subroutine set_fetch_value_descriptor_lim(sbr, io, lim)
C     Set the fetch value descriptor limit to lim.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, lim
      
C     External function.
      logical pio_get_debug_mode
      
      fetch_value_descriptor_lim(io) = lim
      
      call check_value_descriptor_invariant(sbr, io)
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *)
     +        '[Pio/Fortran] Setting fetch_value_descriptor_lim to ',
     +        lim
      endif
      
      return
      end
      
      
      logical function get_memory_mapping_clean(io)
C     Get the memory_mapping_clean flag value.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input argument.
      integer io
      
      get_memory_mapping_clean = memory_mapping_clean(io)
      
      return
      end
      
      
      subroutine set_memory_mapping_clean(io, b)
C     Set the memory_mapping_clean flag to b.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      logical b
      integer io
      
      memory_mapping_clean(io) = b
      
      return
      end
      
      
      subroutine check_memory_mapping_clean(sbr, io, good)
C     Check that the memory_mapping_clean value is good.
      
      implicit none
      
      include 'memory_definition.fh'
      
C     Input arguments.
      logical good
      character sbr*(*)
      integer io
      
C     Local variable.
      logical bad
      
C     External function.
      logical get_memory_mapping_clean
      
      bad = get_memory_mapping_clean(io)
      if (bad .neqv. good) then
         call error_wrong_memory_mapping_clean(sbr, io, good, bad)
      endif
      
      return
      end
      
      
      subroutine check_valid_memory_mapping_slot(sbr, io, loc)
C     Check that loc is a valid slot in the memory mapping.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, loc
      
      if ((loc .lt. MIN_MEMORY_MAPPING_LOCATION) .or.
     +     (loc .gt. MAX_MEMORY_MAPPING_LOCATION)) then
         call error_invalid_memory_mapping_access(sbr, io, loc)
      endif
      
      return
      end
      
      
      subroutine rand_store_slot(sbr, io, loc, v)
C     Store v in slot number loc of the memory mapping.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, loc, v
      
      call check_memory_mapping_clean(sbr, io, .False.)
      call check_valid_memory_mapping_slot(sbr, io, loc)
      
      memory_mapping(io, loc) = v
      
      return
      end
      
      
      integer function rand_fetch_slot(sbr, io, loc)
C     Fetch contents of slot number loc of the memory mapping.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, loc
      
      call check_memory_mapping_clean(sbr, io, .True.)
      call check_valid_memory_mapping_slot(sbr, io, loc)
      
      rand_fetch_slot = memory_mapping(io, loc)
      
      return
      end
      
      
      subroutine next_value_next_value_descriptor(sbr, io)
C     Increment value number and set the first address of its descriptor.
      
      implicit none
      
C     Input arguments.
      character sbr*(*)
      integer io
      
C     Local variables.
      integer svn, svd_ptr
      
C     External functions.
      integer get_store_value_number, get_store_value_descriptor_ptr
      
      call incr_store_value_number(sbr, io)
      
      svn = get_store_value_number(io)
      svd_ptr = get_store_value_descriptor_ptr(io)
      call store_communication_value_descriptor_address
     +     (sbr, io, svn, svd_ptr)
      call advance_store_value_descriptor_ptr
     +     (sbr, io, SLOT_VALUE_SIZES_BEGIN)
      
      return
      end
      
      
      subroutine value_descriptor_relative_store_slot(sbr, io, d, v)
C     Store v in slot number (vda(i) + d) in the memory mapping,
C     with i = store_value_number(io)
C     and vda(i) =
C     memory_mapping(io, SLOT_VALUE_DESCRIPTOR_ADDRESSES + i - 1).
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, d, v
      
C     Local variables.
      integer i, loc
      
C     External function.
      logical pio_get_debug_mode
      
      i = store_value_number(io)
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *)
     +        '[Pio/Fortran] store_value_number(', io, ') = ', i
      endif
      loc = memory_mapping(io, (SLOT_VALUE_DESCRIPTOR_ADDRESS(i)))
      
      if (loc .ne. INVALID_NATURAL) then
         call rand_store_slot(sbr, io, loc + d, v)
      else
         call error_unset_value_descriptor_address(sbr, io, i)
      endif
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *)
     +        '[Pio/Fortran] ',
     +        'Setting memory mapping slot ', loc + d, ' to ', v
      endif
      
      return
      end
      
      
      integer function value_descriptor_relative_fetch_slot
     +     (sbr, io, i, d)
C     Fetch contents of slot number (vda(i) + d) of the memory mapping,
C     with vda(i) =
C     memory_mapping(io, SLOT_VALUE_DESCRIPTOR_ADDRESSES + i - 1).
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, i, d

C     Local variable.
      integer loc
      
C     External function.
      integer rand_fetch_slot
      
      loc = memory_mapping(io, (SLOT_VALUE_DESCRIPTOR_ADDRESS(i)))
      
      if (loc .ne. INVALID_NATURAL) then
         value_descriptor_relative_fetch_slot =
     +        rand_fetch_slot(sbr, io, loc + d)
      else
         call error_unset_value_descriptor_address(sbr, io, i)
      endif
      
      return
      end
      
      
      subroutine seq_store_slot(sbr, io, v)
C     Store v in the current slot of the memory mapping and increment
C     the value descriptor pointer.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io, v
      
C     Local variable.
      integer loc
      
C     External function.
      logical pio_get_debug_mode
      
      loc = store_value_descriptor_ptr(io)
      
      call rand_store_slot(sbr, io, loc, v)
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *)
     +        '[Pio/Fortran] ',
     +        'Setting memory mapping slot ', loc, ' to ', v
      endif
      
      call incr_store_value_descriptor_ptr(sbr, io)
      
      return
      end
      
      
      integer function seq_fetch_slot(sbr, io)
C     Fetch contents of the current slot of the memory mapping and
C     increment the value descriptor pointer.
      
      implicit none
      
      include 'memory_mapping_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer io
      
C     External function.
      integer rand_fetch_slot
      
      seq_fetch_slot =
     +     rand_fetch_slot(sbr, io, fetch_value_descriptor_ptr(io))
      call incr_fetch_value_descriptor_ptr(sbr, io)
      
      return
      end
